/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders (corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

%{
#include "algol60.h"
#include "y.tab.h"
%}

%s COMMENT ALGOL_COMMENT

/* Arithmetic Operators */
op_add		"+"
op_sub		"-"
op_mul		"*"
op_div		"/"
op_mod		"%"
op_exp		"**"

log_and		[Aa][Nn][Dd]
log_or		[Oo][Rr]
log_not		[Nn][Oo][Tt]

cmp_lt		"<"
cmp_le		"<="
cmp_eq		"=="
cmp_ne		"!="
cmp_ge		">="
cmp_gt		">"

lparen		"("
rparen		")"
semicolon	";"
period		"."
comma		","
assignment	":="
newline		"\n"
whitespace	" "|"\t"|"\f"|"\r"
perl_comment	"#".*
cpp_comment	"//".*

digit		[0-9]
letter		[a-zA-Z]
sign		("+"|"-")?

identifier	{letter}({letter}|{digit})*

inum		{sign}{digit}+
/*rnum		{digit}+{period}{digit}+*/
rnum		{sign}({digit}*\.{digit}+|{digit}+\.{digit}*)

key_real	"real"
key_integer	"integer"
key_begin	"begin"
key_end		"end"
key_print	"print"
key_comment	"comment"
key_if		"if"
key_then	"then"
key_else	"else"
key_for		"for"
key_step	"step"
key_until	"until"
key_while	"while"
key_do		"do"
key_procedure	"procedure"
key_byval	"byval"
key_byref	"byref"

%%

{perl_comment}	{ /* eat comments */ }
{cpp_comment}	{ /* eat me */ }

"/*".*"*/"		{ /* one line comment */ }
"/*"	 		{ BEGIN COMMENT; }
<COMMENT>"*/"		{ BEGIN 0; }
<COMMENT>.*		{ /* do nothing */ }

{key_comment}		{ BEGIN ALGOL_COMMENT; }
<ALGOL_COMMENT>{semicolon}	{ BEGIN 0; }
<ALGOL_COMMENT>.*	{ /* ZZzzzz */ }

{op_exp}	{ return OP_EXP; }
{op_add}	{ return OP_ADD; }
{op_sub}	{ return OP_SUB; }
{op_mul}	{ return OP_MUL; }
{op_div}	{ return OP_DIV; }
{op_mod}	{ return OP_MOD; }

{lparen}	{ return LPAREN; }
{rparen}	{ return RPAREN; }
{semicolon}	{ return SEMICOLON; }
{comma}		{ return COMMA; }
{whitespace}	{ /* ZZZzzzzz */ }
{newline}	{ linenum++; }
{assignment}	{ return ASSIGNMENT; }

{log_and}	{ return LOG_AND; }
{log_or}	{ return LOG_OR; }
{log_not}	{ return LOG_NOT; }

{cmp_lt}	{ yylval.i_type = CMP_LT; return COMPARISON; }
{cmp_le}	{ yylval.i_type = CMP_LE; return COMPARISON; }
{cmp_eq}	{ yylval.i_type = CMP_EQ; return COMPARISON; }
{cmp_ne}	{ yylval.i_type = CMP_NE; return COMPARISON; }
{cmp_ge}	{ yylval.i_type = CMP_GE; return COMPARISON; }
{cmp_gt}	{ yylval.i_type = CMP_GT; return COMPARISON; }
{key_else}	{ return KEY_ELSE; }
{key_real}	{ return KEY_REAL; }
{key_integer}	{ return KEY_INTEGER; }
{key_begin}	{ return KEY_BEGIN; }
{key_end}	{ return KEY_END; }
{key_print}	{ return KEY_PRINT; }
{key_if}	{ return KEY_IF; }
{key_then}	{ return KEY_THEN; }

{key_for}	{ return KEY_FOR; }
{key_until}	{ return KEY_UNTIL; }
{key_step}	{ return KEY_STEP; }
{key_while}	{ return KEY_WHILE; }
{key_do}	{ return KEY_DO; }
{key_procedure} { return KEY_PROCEDURE; }
{key_byval}	{ return KEY_BYVAL; }
{key_byref}	{ return KEY_BYREF; }

{inum}		{ DBG("INUM: %d [%s]", atoi(yytext), yytext); yylval.i_type = atoi(yytext); return INUM; }
{rnum}		{ yylval.r_type = atof(yytext); return RNUM; }
{identifier}	{
		  struct symtab *id;
		  char *fooi = (char *)calloc(1, strlen(yytext + 12));
		  sprintf(fooi, "%s&%d", yytext, current_block->level);

		  if ((id = symlook(current_block, yytext)) != NULL) {
		    free(fooi);
		    yylval.var_type = id;
		    switch(id->type) {
		      case INTEGER: return ID_INTEGER;
		      case REAL: return ID_REAL;
		      case PROCEDURE: return ID_PROCEDURE;
		      case BOOL: return ID_BOOL;
		      case VOID: fprintf(stderr, "Internal error: void variable '%s' -> '%s'.\n", yytext, fooi); exit(1);
		    }
		  } else {
		    yylval.cp_type = fooi;
		    return ID_UNKNOWN;
		  }
		}

.		{ yyerror("lexer: encountered unknown token: [%s]", yytext); }
