/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders (corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

%{
extern const char *yytext;
#include "algol60.h"

FILE *outfile;

/* Emit debugging 'write' ASM. */
#define DEBUG_WRITE	0
%}

%union {
  long		i_type;
  double	r_type;
  char		*cp_type;
  struct symtab *var_type;
}

%token LPAREN RPAREN SEMICOLON COMMA
%token ID_UNKNOWN ID_REAL ID_INTEGER ID_PROCEDURE ID_BOOL
%token KEY_BEGIN KEY_END KEY_PRINT KEY_IF KEY_THEN KEY_ELSE KEY_STEP KEY_WHILE KEY_UNTIL KEY_FOR KEY_DO
%token INUM RNUM KEY_REAL KEY_INTEGER KEY_PROCEDURE KEY_BYVAL KEY_BYREF KEY_BOOL

%type <var_type>	i_assigndecl r_assigndecl decl_int decl_real
%type <var_type>	ID_REAL ID_INTEGER ID_PROCEDURE expr assignment procedure_call variable
%type <cp_type>		ID_UNKNOWN
%type <i_type>		INUM COMPARISON type proc_type call_by
%type <r_type>		RNUM

%right		ASSIGNMENT
%left		COMPARISON
%left 		LOG_AND LOG_OR
%left		OP_ADD OP_SUB
%left		OP_MUL OP_DIV OP_MOD
%left		OP_EXP
%nonassoc	LOG_NOT
%nonassoc	UNARY 

%start program

%%

declaration	: KEY_INTEGER decl_int
		| KEY_REAL decl_real
		;

decl_int	: ID_UNKNOWN i_assigndecl i_decllist		{ DBG("declaration: decl int [%s]", $1); asm_decl($1, INTEGER, $2); }
		| ID_INTEGER i_assigndecl i_decllist		{ asm_decl_shadow($1, INTEGER, $2); }
		| ID_REAL i_assigndecl i_decllist		{ asm_decl_shadow($1, INTEGER, $2); }
		| ID_PROCEDURE					{ yyerror("attempt to declare already used procedure '%s' as integer", unamp($1->name)); }
		;

decl_real	: ID_UNKNOWN r_assigndecl r_decllist		{ DBG("declaration: decl real [%s]", $1); asm_decl($1, REAL, $2); }
		| ID_REAL r_assigndecl r_decllist		{ asm_decl_shadow($1, REAL, $2); }
		| ID_INTEGER r_assigndecl r_decllist		{ asm_decl_shadow($1, REAL, $2); }
		| ID_PROCEDURE					{ yyerror("attempt to declare already used procedure '%s' as real", unamp($1->name)); }
		;

i_decllist	: COMMA decl_int				{ /* Nothing for a comma in declaration. */ }
		|
		;

r_decllist	: COMMA decl_real				{ /* Nothing for a comma in declaration. */ }
		|
		;

i_assigndecl	: ASSIGNMENT expr		{ DBG("i_assigndecl: [%s]", $2->name);
						  if ($2->type != INTEGER)
						    yyerror("attempt to assign real value '%s' to integer variable.", unamp($2->name));
						  else
						    $$ = $2;
						}
		|				{ DBG("i_assigndecl: empty"); $$ = symiliteral(0);  }
		;

r_assigndecl	: ASSIGNMENT expr		{ DBG("r_assigndecl: [%s]", $2->name);
						  if ($2->type == INTEGER)
						    $$ = asm_int2real($2);
						  else
						    $$ = $2;
						}
		|				{ DBG("r_assigndecl: empty"); $$ = symrliteral(0.0); }
		;

/* Assignments */

assignment	: ID_INTEGER ASSIGNMENT expr		{ DBG("assignment: [%s] [%s]", $1->name, $3->name); $$ = asm_assign($1, $3); }
		| ID_REAL ASSIGNMENT expr		{ DBG("r_assignment: real [%s] [%s]", $1->name, $3->name); $$ = asm_assign($1, $3); }
		| INUM ASSIGNMENT			{ yyerror("cannot assign to the number '%ld'.", $1); $$ = NULL; }
		| RNUM ASSIGNMENT			{ yyerror("cannot assign to the number '%f'.", $1); $$ = NULL; }
		| ID_UNKNOWN ASSIGNMENT			{ yyerror("cannot assign to undeclared variable '%s'", unamp($1)); $$ = NULL; }
		| ID_PROCEDURE ASSIGNMENT expr		{ $$ = asm_assign($1, $3); }
		;

/* Functions */

proc_decl_head	: procedure_decl proc_decl_list
		;

proc_decl_list	: procedure_decl proc_decl_list
		|
		;

procedure_decl	: KEY_PROCEDURE proc_type ID_UNKNOWN		{ asm_procedure_decl($2, $3); }
		  LPAREN { asm_proc_decl_param(0, NULL, 0, $3); /* print decl label */ }
		  proc_decl_param { /* jump next decl label */ } RPAREN
		  block						{ asm_procedure_decl_cleanup($3); }
		| KEY_PROCEDURE ID_PROCEDURE			{ yyerror("attempt to redefine function '%s'.", unamp($2->name)); }
		;

procedure_call	: ID_PROCEDURE LPAREN	{ asm_proc_call_param_check($1, NULL); }
		  proc_call_param	{ asm_proc_call_param_check(NULL, NULL); }
		  RPAREN		{ $$ = asm_procedure_call($1); }
		;

proc_call_param	: proc_call_many
		|
		;

proc_call_many	: expr					{ asm_proc_call_param_check(NULL, $1); }
		| proc_call_many COMMA expr		{ asm_proc_call_param_check(NULL, $3); }
		;

proc_decl_param	: type ID_UNKNOWN call_by 			{ asm_proc_decl_param($1, $2, $3, NULL); }
		| proc_decl_param COMMA type ID_UNKNOWN call_by { asm_proc_decl_param($3, $4, $5, NULL); }
		|
		;

call_by		 : KEY_BYREF				{ $$ = BYREF; }
		 | KEY_BYVAL				{ $$ = BYVAL; }
		 |					{ $$ = BYVAL; }
		 ;

/* Statements */

program		: block			{ fputs("halt\n", outfile); }
		| 			{ fputs("jmp :start&\n", outfile); }
		  proc_decl_head	{ fputs(":start&\n", outfile); symdeclparam(); }
		  block			{ fputs("halt\n", outfile); }
		| error			{ yyerror("syntaxi ierror"); }
		;

block		: KEY_BEGIN { blockpush(); } statement_list KEY_END { blockpop(); }
		;

statement_list	: statement_list statement
		|
		;

statement	: block
		| declaration SEMICOLON
		| assignment SEMICOLON
		| print_statement SEMICOLON
		| procedure_call SEMICOLON
		| if_statement
		| for_statement
		;

print_statement	: KEY_PRINT expr			{ fprintf(outfile, "write %s\n", $2->name); }
		;

for_statement	:  KEY_FOR assignment KEY_DO block	/* We handle it natively. */
		| /* See below. */
		KEY_FOR assignment KEY_WHILE			{ fprintf(outfile, "%s\n", labelpush()); }
		expr KEY_DO					{ fprintf(outfile, "jf %s,%s\n", $5->name, labelpush()); }
		block						{
								  char labeltst[32], labelout[32];
								  strcpy(labelout, labelpop());
								  strcpy(labeltst, labelpop());
								  fprintf(outfile, "jmp %s\n", labeltst);
								  fprintf(outfile, "%s\n", labelout);
								}
		| KEY_FOR assignment KEY_STEP expr KEY_UNTIL	{ fprintf(outfile, "%s\n", labelpush()); }
		expr KEY_DO					{ fprintf(outfile, "jt %s,%s\n", $7->name, labelpush()); }
		block						{
								  struct symtab *tmp;
								  char labeltst[32], labelout[32];
								  strcpy(labelout, labelpop());
								  strcpy(labeltst, labelpop());

								  tmp = asm_math(MOP_ADD, $4, $2);
								  asm_assign($2, tmp); 
								  fprintf(outfile, "jmp %s\n", labeltst);
								  fprintf(outfile, "%s\n", labelout);
								}
		// Error cases.
		| KEY_FOR assignment block					{ yyerror("missing 'do' keyword in 'for'"); }
/*
		| KEY_FOR assignment error block				{ yyerror("missing 'while' keyword in 'for..while'"); }
		| KEY_FOR assignment KEY_WHILE expr block			{ yyerror("missing 'do' keyword in 'for..while'"); }
		| KEY_FOR assignment KEY_STEP expr KEY_UNTIL expr block		{ yyerror("missing 'do' keyword in 'for..step..until'"); }
		| KEY_FOR assignment KEY_STEP error KEY_DO block		{ yyerror("missing 'until' keyword in 'for..step..until'"); }
		| error { yyerror("error"); }
*/
		;

if_statement	: KEY_IF expr KEY_THEN						{ fprintf(outfile, "jf %s, %s\n", $2->name, labelpush()); }
		if_tail 
		;

if_tail		: block								{ fprintf(outfile, "%s\n", labelpop()); }
		| block 							{
										  const char *labelout = labelpop();
										  fprintf(outfile, "jmp %s\n", labelpush());
										  fprintf(outfile, "%s\n", labelout); 
										}
		KEY_ELSE block							{
										  fprintf(outfile, "%s\n", labelpop());
										}


// The almighty Expression.

variable	: ID_REAL				{ $$ = $1; }
		| ID_INTEGER				{ $$ = $1; }
		;

type		: KEY_REAL				{ $$ = REAL; }
		| KEY_INTEGER				{ $$ = INTEGER; }
		| KEY_BOOL				{ $$ = BOOL; }
		| KEY_PROCEDURE				{ $$ = PROCEDURE; }
		;

proc_type	: type					{ $$ = $1; }
		|					{ $$ = VOID; }
		;

expr		: INUM					{ DBG("INUM: literal to var: %ld", $1); $$ = symiliteral($1); }
		| RNUM					{ DBG("RNUM: literal to var: %f", $1); $$ = symrliteral($1); }
		| ID_UNKNOWN				{ yyerror("attempt to use undeclared variable '%s'", unamp($1)); }
		| ID_PROCEDURE				{ yyerror("attempt to use procedure '%s' as a variable, use '%s()'.", unamp($1->name), unamp($1->name)); }
		| variable				{ $$ = $1; }
		| procedure_call			{ if ($1->type == VOID)
							    yyerror("attempt to use non-typed function '%s' in expression", unamp($1->name));
							  else
							    $$ = $1;
							}
		// Assignments
		| assignment				{ $$ = $1; }
		// Logical
		| expr LOG_AND expr			{ $$ = asm_log("and", $1, $3); }
		| expr LOG_OR expr			{ $$ = asm_log("or", $1, $3); }
		| LOG_NOT expr				{ $$ = asm_log("not", $2, NULL); }
		// Mathematical
		| OP_SUB expr 		%prec UNARY	{ struct symtab *zero = symiliteral(0); $$ = asm_math(MOP_SUB, zero, $2); }
		| expr OP_ADD expr			{ $$ = asm_math(MOP_ADD, $1, $3); }
		| expr OP_SUB expr			{ $$ = asm_math(MOP_SUB, $1, $3); }
		| expr OP_MUL expr			{ $$ = asm_math(MOP_MUL, $1, $3); }
		| expr OP_DIV expr			{ $$ = asm_math(MOP_DIV, $1, $3); }
		| expr OP_MOD expr			{ $$ = asm_math(MOP_MOD, $1, $3); }
		| expr OP_EXP expr			{ $$ = asm_math_exp($1, $3); }
		| LPAREN expr RPAREN            	{ $$ = $2; }
		// Comparisons
		| expr COMPARISON expr			{ $$ = asm_cmp($2, $1, $3); }
		// Errors
		| LPAREN RPAREN				{ yyerror("empty expression"); }
		;
%%

int yyparse(void);
int yylex(void);

int linenum = 1;

int main()
{
  outfile = stdout;
  blockinit();
  syminit(current_block);
  labelinit();
  refparaminit();
  yyparse();
  return 0;
}
