/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders <corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include "algol60.h"

struct symtab *asm_decl_shadow(struct symtab *current, enum e_type type, struct symtab *value)
{
  char *newvar, *oldvar, *here;

  if (!symdeclok(current_block, current))
    yyerror("attempt to declare already used %s '%s' as %s", symtype(current->type), unamp(current->name), symtype(type));

  newvar = calloc(1, strlen(current->name) + 12);
  oldvar = strdup(current->name);
  if ((here = strrchr(oldvar, '&')))
    *here = '\0';
  sprintf(newvar, "%s&%d", oldvar, current_block->level);
  free(oldvar);

  return asm_decl(newvar, type, value);
}

struct symtab *asm_decl(char *varname, enum e_type type, struct symtab *value)
{
  struct symtab *myvar;

  myvar = syminsert(current_block, varname, BYVAL, type);
  fprintf(outfile, "declare %s,%s\n", varname, symtype(type));
  asm_assign(myvar, value);
  return myvar;
}

struct symtab *asm_assign(struct symtab *assignto, struct symtab *assignfrom)
{
  if (!assignto || !assignfrom)
    yyerror("null values to asm_assign");

  if (assignto->type != PROCEDURE) {
    if (assignto->type == INTEGER && assignfrom->type == REAL)
      yyerror("attempt to assign real expression '%s' to integer variable '%s'.", unamp(assignfrom->name), unamp(assignto->name));
  //    assignfrom = asm_real2int(assignfrom);
    else if (assignto->type == REAL && assignfrom->type == INTEGER)
      assignfrom = asm_int2real(assignfrom);
   
    fprintf(outfile, "store %s,%s\n", assignfrom->name, assignto->name);
    return assignto;
  } else {
    char *procret = proc_param_name(assignto->name, 0);
    fprintf(outfile, "store %s,%s\n", assignfrom->name, procret);
    free(procret);
    return symlook(root_block, procret);
  }
  yyerror("Hi mom!");
  return NULL;
}

void asm_assign_real(struct symtab *symvar, float rval)
{
  if (!symvar)
    yyerror("attempt to assign to non-existent variable");

  if (symvar->type != REAL)
    yyerror("assigning a real number to non-real variable '%s'", unamp(symvar->name));

  /* ASM */
  fprintf(outfile, "store %f,%s\n", rval, symvar->name);
}

void asm_assign_int(struct symtab *symvar, long ival)
{
  if (!symvar)
    yyerror("attempt to assign to non-existent variable");

  if (symvar->type != INTEGER)
    yyerror("assigning an integer to non-integer variable '%s'", unamp(symvar->name));

  /* ASM */
  fprintf(outfile, "store %ld,%s\n", ival, symvar->name);
}
