/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders <corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include "algol60.h"

struct compare_operator {
  enum cmp_t token;
  const char *asm_op;
} cmptab[] = {
  { CMP_GE, "high" },
  { CMP_GT, "high" },
  { CMP_EQ, "equ" },
  { CMP_NE, "equ" },
  { CMP_LT, "low" },
  { CMP_LE, "low" },
  { -1, NULL }
};

struct symtab *asm_cmp(enum cmp_t token, struct symtab *a, struct symtab *b)
{
  int i;

//  fprintf(stderr, "asm_cmp: [%s] [%s]\n", a->name, b->name);

  for (i = 0; cmptab[i].token >= 0; i++)
    if (cmptab[i].token == token)
      break;

  switch (token) {
  case CMP_GE:
  case CMP_LE: {
      struct symtab *r2, *r3;
      fprintf(outfile, "%s %s,%s,%s\n", cmptab[i].asm_op, a->name, b->name, (r2 = symtempvar(INTEGER))->name);
      fprintf(outfile, "equ %s,%s,%s\n", a->name, b->name, (r3 = symtempvar(INTEGER))->name);
      return asm_log("or", r2, r3);
    }

  case CMP_LT:
  case CMP_EQ:
  case CMP_NE:
  case CMP_GT: {
      struct symtab *r1 = symtempvar(INTEGER);
      fprintf(outfile, "%s %s,%s,%s\n", cmptab[i].asm_op, a->name, b->name, r1->name);

      if (token == CMP_NE)
        return asm_log("not", r1, NULL);
      else
        return r1;
    }
  }

  yyerror("asm_cmp: invalid comparison: %d", token);
  return NULL;
}
