/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders <corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include <ctype.h>
#include "algol60.h"

struct {
  enum mathop_t op;
  const char *op_text;
  const char *asm_name;
} asm_math_ops[] = {
  { MOP_ADD, "+", "add" },
  { MOP_SUB, "-", "sub" },
  { MOP_MUL, "*", "mult" },
  { MOP_DIV, "/", "div" },
  { MOP_MOD, "%", "mod" },
  { 0, NULL, NULL }
};

struct symtab *asm_math(enum mathop_t op, struct symtab *a, struct symtab *b)
{
  int i;
  char prefix;
  struct symtab *tmpie;

  if (a->type == b->type && a->type == INTEGER) {
    prefix = 'i';
    tmpie = symtempvar(INTEGER);
  } else {
    if (op == MOP_MOD)
      yyerror("cannot modulus reals");

    if (a->type == INTEGER) {
      struct symtab *convert_var = symtempvar(REAL);
      DBG("asm_math: itor tempvar %s [%p]", convert_var->name, convert_var);
      fprintf(outfile, "itor %s,%s\n", a->name, convert_var->name);
      a = convert_var;
    } else if (b->type == INTEGER) {
      struct symtab *convert_var = symtempvar(REAL);
      DBG("asm_math: itor tempvar %s [%p]", convert_var->name, convert_var);
      fprintf(outfile, "itor %s,%s\n", b->name, convert_var->name);
      b = convert_var;
    }
    prefix = 'r';
    tmpie = symtempvar(REAL);
  }

  DBG("asm_math: tempvar %s [%p]", tmpie->name, tmpie);

  for (i = 0; asm_math_ops[i].op_text != NULL; i++)
    if (asm_math_ops[i].op == op)
      break;

  if (asm_math_ops[i].op_text == NULL)
    yyerror("asm_math: invalid op: %d", op);

  fprintf(outfile, "%c%s %s,%s,%s\n", prefix, asm_math_ops[i].asm_name, a->name, b->name, tmpie->name);
  
  return tmpie;
}

struct symtab *asm_int2real(struct symtab *intval)
{
  struct symtab *tmpie = symtempvar(REAL);
  fprintf(outfile, "itor %s,%s\n", intval->name, tmpie->name);
  return tmpie;
}

struct symtab *asm_real2int(struct symtab *realval)
{
  struct symtab *tmpie = symtempvar(INTEGER);
  fprintf(outfile, "rtoi %s,%s\n", realval->name, tmpie->name);
  return tmpie;
}

struct symtab *asm_math_exp(struct symtab *base, struct symtab *exp)
{
  long times;
  struct symtab *ret;

  if (exp->type == REAL)
    yyerror("Exponentiation by a real number is not supported.");

  if ((times = atoi(exp->name)) == 0 && isdigit(*exp->name))
    return (base->type == REAL ? symrliteral(1.0) : symiliteral(1));
  else if (times == 1)
    return base;

#if 1
{
  struct symtab *itr, *one, *zero, *compare, *itrswap, *retswap;

  one = symiliteral(1);
  zero = symiliteral(0);

  ret = symtempvar(base->type);

  asm_assign(ret, one);
  compare = asm_cmp(CMP_EQ, exp, zero);
  fprintf(outfile, "jt %s,:expout%s\n", compare->name, ret->name);

  asm_assign(ret, base);
  compare = asm_cmp(CMP_EQ, exp, one);
  fprintf(outfile, "jt %s,:expout%s\n", compare->name, ret->name);

  itr = symtempvar(INTEGER);
  asm_assign(itr, exp);

  fprintf(outfile, ":exp%s\n", ret->name);

  retswap = asm_math(MOP_MUL, ret, base);
  asm_assign(ret, retswap);
  itrswap = asm_math(MOP_SUB, itr, one);
  asm_assign(itr, itrswap);
  compare = asm_cmp(CMP_EQ, itr, one);

  fprintf(outfile, "jf %s,:exp%s\n", compare->name, ret->name);
  fprintf(outfile, ":expout%s\n", ret->name);
}
#else
{
  struct symtab copy;

  ret = base;
  memcpy(&copy, base, sizeof(struct symtab));
  while (--times > 0) {
    ret = asm_math(MOP_MUL, ret, &copy);
}
#endif

  return ret;
}
