/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders <corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include "algol60.h"

queue_t * newqueue(void)
{
  queue_t *queue;
  queue = (queue_t *) calloc(1, sizeof(queue_t));
  assert(queue != NULL);

  queue->nelems = 0;
  queue->root = NULL;
  queue->current = queue->root;

  return queue;
}

void deletequeue(queue_t *queue)
{
  assert(queue != NULL);
  while( !qisempty(queue) )
  	dequeue(queue);
  free(queue);
}

void enqueue(queue_t *queue, void *elem)
{
  node_t *next;
  next = (node_t *) calloc(1, sizeof(node_t));
  assert(next != NULL);

  next->prev = NULL;
  next->data = elem;

  if( queue->nelems == 0 )
    queue->current = queue->root = next;
  else 
    queue->root->prev = next;

  queue->nelems++;
}

void * dequeue(queue_t *queue)
{
  node_t *junk;
  void *elem;

  junk = queue->current;
  assert(junk != NULL);

  elem = junk->data;
  queue->current = queue->current->prev;

  free(junk);
  queue->nelems--;

  if(queue->nelems == 0) { queue->root = queue->current = NULL; }

  return elem;
}

int qisempty(queue_t *queue)
{
  return (!queue->nelems);
}

int qlength(queue_t *queue)
{
  return queue->nelems;
}
