/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders <corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include "algol60.h"

stack_t * newstack(void)
{
  stack_t *stack;
  stack = (stack_t *) calloc(1, sizeof(stack_t));
  assert(stack != NULL);

  stack->nelems = 0;
  stack->root = NULL;
  stack->current = stack->root;

  return stack;
}

void deletestack(stack_t *stack)
{
  assert(stack != NULL);
  while( !isempty(stack) )
  	pop(stack);
  free(stack);
}

void push(stack_t *stack, void *elem)
{
  node_t *next;
  next = (node_t *) calloc(1, sizeof(node_t));
  assert(next != NULL);

  next->prev = stack->current;
  next->data = elem;

  stack->current = next;
  stack->nelems++;
}

void * pop(stack_t *stack)
{
  node_t *junk;
  void *elem;

  junk = stack->current;
  assert(junk != NULL);

  elem = junk->data;
  stack->current = stack->current->prev;

  free(junk);
  stack->nelems--;

  return elem;
}

void * peek(stack_t *stack)
{
  if( !stack->nelems ) return NULL;
  
  return stack->current->data;
}

int isempty(stack_t *stack)
{
  return (!stack->nelems);
}

int length(stack_t *stack)
{
  return stack->nelems;
}
